from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import warnings
from keras.callbacks import Callback
from keras.models import Model

class ModelCheckpoint(Callback):

	def __init__(self, filepath, base_model, monitor='val_loss', verbose=0,
				 save_best_only=False, save_weights_only=False,
				 mode='auto', period=1):
		super(ModelCheckpoint, self).__init__()
		self.base_model = base_model
		self.monitor = monitor
		self.verbose = verbose
		self.filepath = filepath
		self.save_best_only = save_best_only
		self.save_weights_only = save_weights_only
		self.period = period
		self.epochs_since_last_save = 0

		if mode not in ['auto', 'min', 'max']:
			warnings.warn('ModelCheckpoint mode %s is unknown, '
						  'fallback to auto mode.' % (mode),
						  RuntimeWarning)
			mode = 'auto'

		if mode == 'min':
			self.monitor_op = np.less
			self.best = np.Inf
		elif mode == 'max':
			self.monitor_op = np.greater
			self.best = -np.Inf
		else:
			if 'acc' in self.monitor or self.monitor.startswith('fmeasure'):
				self.monitor_op = np.greater
				self.best = -np.Inf
			else:
				self.monitor_op = np.less
				self.best = np.Inf

	def on_epoch_end(self, epoch, logs=None):
		logs = logs or {}
		self.epochs_since_last_save += 1
		if self.epochs_since_last_save >= self.period:
			self.epochs_since_last_save = 0
			filepath = self.filepath.format(epoch=epoch + 1, **logs)
			if self.save_best_only:
				current = logs.get(self.monitor)
				if current is None:
					warnings.warn('Can save best model only with %s available, '
								  'skipping.' % (self.monitor), RuntimeWarning)
				else:
					if self.monitor_op(current, self.best):
						if self.verbose > 0:
							print('\nEpoch %05d: %s improved from %0.5f to %0.5f,'
								  ' saving model to %s'
								  % (epoch + 1, self.monitor, self.best,
									 current, filepath))
						self.best = current
						if self.save_weights_only:
							self.save_model = Model(inputs=[self.base_model.layers[0].input], outputs=[self.model.layers[-1].output, self.base_model.layers[-1].output])
							self.save_model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
							self.save_model.save_weights(filepath, overwrite=True)
						else:
							self.save_model = Model(inputs=[self.base_model.layers[0].input], outputs=[self.model.layers[-1].output, self.base_model.layers[-1].output])
							self.save_model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
							self.save_model.save(filepath, overwrite=True)
					else:
						if self.verbose > 0:
							print('\nEpoch %05d: %s did not improve from %0.5f' %
								  (epoch + 1, self.monitor, self.best))
			else:
				if self.verbose > 0:
					print('\nEpoch %05d: saving model to %s' % (epoch + 1, filepath))
				if self.save_weights_only:
					self.save_model = Model(inputs=[self.base_model.layers[0].input], outputs=[self.model.layers[-1].output, self.base_model.layers[-1].output])
					self.save_model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
					self.save_model.save_weights(filepath, overwrite=True)
				else:
					self.save_model = Model(inputs=[self.base_model.layers[0].input], outputs=[self.model.layers[-1].output, self.base_model.layers[-1].output])
					self.save_model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
					self.save_model.save(filepath, overwrite=True)